@@ -11,7 +11,6 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import functools
 import sys
 import types
 
@@ -20,10 +19,8 @@ from oslo_log import log as logging
 from oslo_utils import encodeutils
 from oslo_utils import excutils
 import six
-from six import reraise as raise_
 
 from heat.common.i18n import _
-from heat.common.i18n import _LI
 from heat.common.i18n import repr_wrapper
 from heat.common import timeutils
 
@@ -49,7 +46,6 @@ def task_description(task):
     return encodeutils.safe_decode(repr(task))
 
 
-@functools.total_ordering
 class Timeout(BaseException):
     """Raised when task has exceeded its allotted (wallclock) running time.
 
@@ -80,14 +76,11 @@ class Timeout(BaseException):
             generator.close()
             return False
 
-    def __eq__(self, other):
-        if not isinstance(other, Timeout):
-            return NotImplemented
-        return not (self < other or other < self)
+    def earlier_than(self, other):
+        if other is None:
+            return True
 
-    def __lt__(self, other):
-        if not isinstance(other, Timeout):
-            return NotImplemented
+        assert isinstance(other, Timeout), "Invalid type for Timeout compare"
         return self._duration.endtime() < other._duration.endtime()
 
 
@@ -154,7 +147,7 @@ class TaskRunner(object):
     def _sleep(self, wait_time):
         """Sleep for the specified number of seconds."""
         if ENABLE_SLEEP and wait_time is not None:
-            LOG.debug('%s sleeping' % six.text_type(self))
+            LOG.debug('%s sleeping', six.text_type(self))
             eventlet.sleep(wait_time)
 
     def __call__(self, wait_time=1, timeout=None, progress_callback=None):
@@ -181,7 +174,7 @@ class TaskRunner(object):
         assert self._runner is None, "Task already started"
         assert not self._done, "Task already cancelled"
 
-        LOG.debug('%s starting' % six.text_type(self))
+        LOG.debug('%s starting', six.text_type(self))
 
         if timeout is not None:
             self._timeout = Timeout(self, timeout)
@@ -193,7 +186,7 @@ class TaskRunner(object):
         else:
             self._runner = False
             self._done = True
-            LOG.debug('%s done (not resumable)' % six.text_type(self))
+            LOG.debug('%s done (not resumable)', six.text_type(self))
 
     def step(self):
         """Run another step of the task.
@@ -208,18 +201,18 @@ class TaskRunner(object):
                 return False
 
             if self._timeout is not None and self._timeout.expired():
-                LOG.info(_LI('%s timed out'), six.text_type(self))
+                LOG.info('%s timed out', self)
                 self._done = True
 
                 self._timeout.trigger(self._runner)
             else:
-                LOG.debug('%s running' % six.text_type(self))
+                LOG.debug('%s running', six.text_type(self))
 
                 try:
                     poll_period = next(self._runner)
                 except StopIteration:
                     self._done = True
-                    LOG.debug('%s complete' % six.text_type(self))
+                    LOG.debug('%s complete', six.text_type(self))
                 else:
                     if isinstance(poll_period, six.integer_types):
                         self._poll_period = max(poll_period, 1)
@@ -247,7 +240,7 @@ class TaskRunner(object):
         else:
             if timeout is not None:
                 new_timeout = Timeout(self, timeout)
-                if self._timeout is None or new_timeout < self._timeout:
+                if new_timeout.earlier_than(self._timeout):
                     self._timeout = new_timeout
 
         done = self.step() if resuming else self.done()
@@ -277,13 +270,13 @@ class TaskRunner(object):
             return
 
         if not self.started() or grace_period is None:
-            LOG.debug('%s cancelled' % six.text_type(self))
+            LOG.debug('%s cancelled', six.text_type(self))
             self._done = True
             if self.started():
                 self._runner.close()
         else:
             timeout = TimedCancel(self, grace_period)
-            if self._timeout is None or timeout < self._timeout:
+            if timeout.earlier_than(self._timeout):
                 self._timeout = timeout
 
     def started(self):
@@ -421,47 +414,50 @@ class DependencyTaskGroup(object):
         raised_exceptions = []
         thrown_exceptions = []
 
-        while any(six.itervalues(self._runners)):
-            try:
-                for k, r in self._ready():
-                    r.start()
-                    if not r:
-                        del self._graph[k]
+        try:
+            while any(six.itervalues(self._runners)):
+                try:
+                    for k, r in self._ready():
+                        r.start()
+                        if not r:
+                            del self._graph[k]
 
-                if self._graph:
+                    if self._graph:
+                        try:
+                            yield
+                        except Exception:
+                            thrown_exceptions.append(sys.exc_info())
+                            raise
+
+                    for k, r in self._running():
+                        if r.step():
+                            del self._graph[k]
+                except Exception:
+                    exc_info = None
                     try:
-                        yield
-                    except Exception:
-                        thrown_exceptions.append(sys.exc_info())
-                        raise
+                        exc_info = sys.exc_info()
+                        if self.aggregate_exceptions:
+                            self._cancel_recursively(k, r)
+                        else:
+                            self.cancel_all(grace_period=self.error_wait_time)
+                        raised_exceptions.append(exc_info)
+                    finally:
+                        del exc_info
+                except:  # noqa
+                    with excutils.save_and_reraise_exception():
+                        self.cancel_all()
 
-                for k, r in self._running():
-                    if r.step():
-                        del self._graph[k]
-            except Exception:
-                exc_info = sys.exc_info()
-                if self.aggregate_exceptions:
-                    self._cancel_recursively(k, r)
-                else:
-                    self.cancel_all(grace_period=self.error_wait_time)
-                raised_exceptions.append(exc_info)
-                del exc_info
-            except:  # noqa
-                with excutils.save_and_reraise_exception():
-                    self.cancel_all()
-
-        if raised_exceptions:
-            try:
+            if raised_exceptions:
                 if self.aggregate_exceptions:
                     raise ExceptionGroup(v for t, v, tb in raised_exceptions)
                 else:
                     if thrown_exceptions:
-                        raise_(*thrown_exceptions[-1])
+                        six.reraise(*thrown_exceptions[-1])
                     else:
-                        raise_(*raised_exceptions[0])
-            finally:
-                del raised_exceptions
-                del thrown_exceptions
+                        six.reraise(*raised_exceptions[0])
+        finally:
+            del raised_exceptions
+            del thrown_exceptions
 
     def cancel_all(self, grace_period=None):
         if callable(grace_period):
@@ -471,14 +467,20 @@ class DependencyTaskGroup(object):
                 return grace_period
 
         for k, r in six.iteritems(self._runners):
-            gp = get_grace_period(k)
+            if not r.started() or r.done():
+                gp = None
+            else:
+                gp = get_grace_period(k)
             try:
                 r.cancel(grace_period=gp)
             except Exception as ex:
-                LOG.debug('Exception cancelling task: %s' % six.text_type(ex))
+                LOG.debug('Exception cancelling task: %s', six.text_type(ex))
 
     def _cancel_recursively(self, key, runner):
-        runner.cancel()
+        try:
+            runner.cancel()
+        except Exception as ex:
+            LOG.debug('Exception cancelling task: %s', six.text_type(ex))
         node = self._graph[key]
         for dependent_node in node.required_by():
             node_runner = self._runners[dependent_node]
